import os
import pdftotext

"""
## PDF파일을 Text로 변환 ##
- file_name, pdf_text, update_idx 딕셔너리들을 리스트로 반환
디비 적재시 원하는 키값을 확인

- update 쿼리 작성시 유의사항
다운로드 파일 제일 앞에 where문으로 검색할 인덱스값을 붙어야함
ex. (where문으로 컬럼에서 찾을 인덱스)_파일명.pdf
"""


class PdftoText:
    def pdf_to_text(self, download_path):
        pdf_list = []
        for file in os.listdir(download_path):
            if not file.endswith(".pdf"):
                continue

            pdf_path = os.path.join(download_path, file)
            if os.path.getsize(pdf_path) == 0:
                os.remove(pdf_path)
                continue

            # pdf -> text 변환
            with open(pdf_path, "rb") as f:
                pdf = pdftotext.PDF(f)

            pdf_text = ""
            for page in pdf:
                pdf_text += page

            if len(pdf_text) == 0:
                os.remove(pdf_path)
                continue

            # 디비에 update시 where문으로 찾기위한 인덱스
            update_idx = file.split("_")[0]

            data = {
                "file_name": str(file),
                "pdf_text": pdf_text,
                "update_idx": str(update_idx),
            }
            pdf_list.append(data)
        return pdf_list



import os
from typing import Any, List, Tuple

import psycopg2
from dotenv import find_dotenv, load_dotenv
from loguru import logger
from sqlalchemy import Column, MetaData, Table, create_engine
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine import URL, Engine

load_dotenv(dotenv_path=find_dotenv())


def __get_db_info(target: str):
    host = os.environ.get(f"{target}_DB_HOST")
    port = os.environ.get(f"{target}_DB_PORT")
    database = os.environ.get(f"{target}_DB_NAME")
    username = os.environ.get(f"{target}_DB_USERNAME")
    password = os.environ.get(f"{target}_DB_PASSWORD")

    if None in (host, port, database, username, password):
        raise Exception(f"{target}에 대한 환경변수를 제대로 입력해주세요.")

    return host, port, database, username, password


def get_conn(target: str):
    try:
        host, port, database, username, password = __get_db_info(target=target)

        return psycopg2.connect(
            host=host,
            port=port,
            database=database,
            user=username,
            password=password,
        )
    except Exception as error:
        logger.error(error)


def get_engine(target: str, drivername: str = "postgresql+psycopg2"):
    try:
        host, port, database, username, password = __get_db_info(target=target)

        dsn = URL.create(
            host=host,
            port=port,
            database=database,
            username=username,
            password=password,
            drivername=drivername,
        )
        return create_engine(dsn)
    except Exception as error:
        logger.error(error)


def create_to_postgres(
    engine: Engine, name: str, *columns: Tuple[Column[Any]], metadata=MetaData()
):
    """
    예시 코드:

    create_to_postgres(
        get_engine("TARGET_DB"),
        "users",
        Column("id", Integer, primary_key=True),
        Column("name", String),
        Column("age", Integer),
        Column("contact", String),
        UniqueConstraint("name", "contact"),
    )
    """

    try:
        new_table = Table(name, metadata, *columns)
        new_table.create(engine, checkfirst=True)
    except Exception as error:
        logger.error(error)


def insert_to_postgres(
    engine: Engine,
    name: str,
    values: List[dict],
    metadata=MetaData(),
    index_elements=None,
):
    """
    예시 코드:

    users = [
        {"name": "허해준", "age": "29", "contact": "01012345678"},
        {"name": "권재선", "age": "30", "contact": "01087654321"},
    ]
    insert_to_postgres(get_engine("TARGET_DB"), "users", users, index_elements=["name", "contact"])
    """

    try:
        metadata.bind = engine
        table = Table(name, metadata, autoload_with=engine)

        stmt = (
            insert(table)
            .values(values)
            .on_conflict_do_nothing(index_elements=index_elements)
        )

        with engine.begin() as connection:
            connection.execute(stmt)
    except Exception as error:
        logger.error(error)



import pandas as pd
from sqlalchemy import text


def reset_key_active(engine):
    try:
        query = """
                UPDATE t_auth_naver_datalab_key_v2
                SET is_active          = FALSE,
                    failed_active_date = NULL
                WHERE NOT is_active
                  AND failed_active_date < CURRENT_DATE;
                """

        with engine.begin() as connection:
            connection.execute(text(query))

        return True
    except Exception as error:
        raise Exception(error)


def get_key(engine):
    try:
        if engine is None:
            raise Exception("마대리 인증 정보 DB에 접근할 수 없습니다.")

        reset_key_active(engine)

        query = """
              SELECT id, client_id, client_secret
              FROM t_auth_naver_datalab_key_v2
              WHERE is_active
                AND is_auth
              ORDER BY RANDOM()
              LIMIT 1;
              """
        df = pd.read_sql(query, engine)

        if df.empty:
            return None, None, None

        return df.loc[0, "id"], df.loc[0, "client_id"], df.loc[0, "client_secret"]
    except Exception as error:
        raise Exception(error)


def update_key(pk_id, feather, engine):
    try:
        if engine is None:
            raise Exception("마대리 인증 정보 DB에 접근할 수 없습니다.")

        if feather not in ["auth", "active"]:
            raise Exception("사용할 수 없는 feather입니다.")

        query = f"""
                UPDATE t_auth_naver_datalab_key_v2
                SET is_{feather} = FALSE,
                    failed_{feather}_date = CURRENT_DATE
                WHERE id = {pk_id}
                """

        with engine.begin() as connection:
            connection.execute(text(query))

        return True
    except Exception as error:
        raise Exception(error)



import base64
import hashlib
import hmac
import json
from typing import List, TypedDict

import pendulum
import requests
from loguru import logger
from sqlalchemy import text

from dmk_packages.database import database as db

tz = pendulum.timezone("Asia/Seoul")

# fmt: off
DatalabData = TypedDict("Data", {"period": str, "ratio": float})
DatalabResult = TypedDict("DatalabData", {"keyword": str, "data": DatalabData})
SearchadResult = TypedDict("SearchadResult", {"keyword": str, "monthlyPcQcCnt": int, "monthlyMobileQcCnt": int})
# fmt: on


class NaverSearchVolumeKeyHandler:
    def __init__(self, target="MADERI_AUTH"):
        self._engine = db.get_engine(target)
        if not self._engine:
            raise ValueError("데이터베이스 엔진 설정에 실패했습니다.")

        self.__reset_datalab_key()
        self._datalab_key = self._get_datalab_key()
        self._searchad_key = self._get_searchad_key()

    def __reset_datalab_key(self):
        try:
            query = """
            UPDATE t_auth_naver_datalab_key_v2
            SET is_active          = TRUE,
                failed_active_date = NULL
            WHERE NOT is_active
              AND failed_active_date < CURRENT_DATE;
            """

            with self._engine.begin() as connection:
                connection.execute(text(query))
        except Exception as err:
            logger.error(err)

    def _get_searchad_key(self):
        try:
            query = """
            SELECT id, customer_id, access_license, private_key
            FROM t_auth_naver_searchad_key_v2
            WHERE is_valid
            ORDER BY RANDOM()
            LIMIT 1;
            """

            with self._engine.begin() as connection:
                result = connection.execute(text(query))

            searchad_key = result.fetchone()

            if not searchad_key:
                raise ValueError("사용 가능한 검색광고 키가 없습니다.")

            return {
                "key_id": searchad_key[0],
                "customer_id": searchad_key[1],
                "access_license": searchad_key[2],
                "private_key": searchad_key[3],
            }
        except Exception as err:
            logger.error(err)

    def _get_datalab_key(self):
        try:
            query = """
            SELECT id, client_id, client_secret
            FROM t_auth_naver_datalab_key_v2
            WHERE is_auth AND is_active
            ORDER BY RANDOM()
            LIMIT 1;
            """

            with self._engine.begin() as connection:
                result = connection.execute(text(query))

            datalab_key = result.fetchone()

            if not datalab_key:
                raise ValueError("사용 가능한 데이터랩 키가 없습니다.")

            return {
                "key_id": datalab_key[0],
                "client_id": datalab_key[1],
                "client_secret": datalab_key[2],
            }
        except Exception as err:
            logger.error(err)

    # NOTE: 검색광고 API에 대해서는 제한량이 없다고 하여 사용이 안될 수 있다.
    #     : 하지만 만약 그외의 이유로 키를 사용하지 못하게 될 경우를 대비해야한다.
    def _update_searchad_key(self, pk_id, feather):
        try:
            if self._engine is None:
                raise Exception("인증 정보 DB에 접근할 수 없습니다.")

            if feather not in ["valid"]:
                raise Exception("사용할 수 없는 feather입니다.")

            query = f"""
                    UPDATE t_auth_naver_searchad_key_v2
                    SET is_{feather} = FALSE, failed_{feather}_at = CURRENT_DATE
                    WHERE id = {pk_id};
                    """

            with self._engine.begin() as connection:
                connection.execute(text(query))

        except Exception as error:
            raise Exception(error)

    def _update_datalab_key(self, pk_id, feather):
        try:
            if self._engine is None:
                raise Exception("인증 정보 DB에 접근할 수 없습니다.")

            if feather not in ["auth", "active"]:
                raise Exception("사용할 수 없는 feather입니다.")

            query = f"""
                    UPDATE t_auth_naver_datalab_key_v2
                    SET is_{feather} = FALSE, failed_{feather}_date = CURRENT_DATE
                    WHERE id = {pk_id};
                    """

            with self._engine.begin() as connection:
                connection.execute(text(query))

        except Exception as error:
            raise Exception(error)


class NaverSearchVolumeCrawler(NaverSearchVolumeKeyHandler):
    DEFAULT_DT = pendulum.yesterday(tz=tz)
    DATALAB_URL = "https://openapi.naver.com/v1/datalab/search"
    SEARCHAD_URL = "https://api.naver.com/keywordstool"

    LESS_THAN_10 = "< 10"
    MPQC = "monthlyPcQcCnt"
    MMQC = "monthlyMobileQcCnt"

    def __init__(self, date_from=None, date_until=None):
        super(NaverSearchVolumeCrawler, self).__init__()

        self._date_range = []
        self._date_from = pendulum.parse(date_from) if date_from else self.DEFAULT_DT
        self._date_until = pendulum.parse(date_until) if date_until else self.DEFAULT_DT

        self._keywords: List[str] | None = None
        self._keywords_bundle: List[List[str]] | None = None

    def _set_date_range(self, target_date):
        try:
            # NOTE: 30일전 날짜를 구해야한다.
            #     : 타겟날짜도 포함하기에 코드에서는 29일을 차감한다.
            date_thirty_days_ago = target_date.subtract(days=29)

            self._date_range.append([date_thirty_days_ago, target_date])

            if self._date_from < date_thirty_days_ago:
                return self._set_date_range(date_thirty_days_ago.subtract(days=1))
        except Exception as error:
            logger.error(error)

    def _set_keywords(self, keywords: List[str]):
        self._keywords = keywords

    def _set_keywords_bundle(self, keywords: List[str]):
        try:
            keywods_bundle = [keywords[i : i + 5] for i in range(0, len(keywords), 5)]
            self._keywords_bundle = keywods_bundle
        except Exception as error:
            logger.error(error)

    def get_keywords(self) -> List[str]:
        """해당 메서드를 오버라이딩하여 키워드를 세팅합니다."""

        # NOTE: 예시를 위한 키워드입니다.
        return ["Python", "Rust", "JavaScript", "Java", "Flutter"]

    def get_searchad_data(self, keywords_bundle: List[str]) -> List[SearchadResult]:
        """
        네이버 검색광고 API 데이터 가져오기

        API 문서: https://naver.github.io/searchad-apidoc/#/tags/RelKwdStat
        """
        try:
            if len(keywords_bundle) > 5:
                raise ValueError("keywords_bundle 리스트는 최대 5개까지만 허용됩니다.")

            hint_keywords = ",".join(
                [k.replace(" ", "").upper() for k in keywords_bundle]
            )
            hint_keywords_map = {k.replace(" ", "").upper(): k for k in keywords_bundle}

            timestamp = str(round(pendulum.now().timestamp() * 1000))
            signature = base64.b64encode(
                hmac.new(
                    bytes(self._searchad_key["private_key"], "UTF-8"),
                    bytes(f"{timestamp}.GET./keywordstool", "UTF-8"),
                    digestmod=hashlib.sha256,
                ).digest()
            )

            headers = {
                "Content-Type": "application/json; charset=UTF-8",
                "X-Timestamp": timestamp,
                "X-API-KEY": self._searchad_key["access_license"],
                "X-Customer": self._searchad_key["customer_id"],
                "X-Signature": signature,
            }

            response = requests.get(
                url=self.SEARCHAD_URL,
                headers=headers,
                params={"hintKeywords": hint_keywords},
            )

            if not response.ok:
                raise Exception(response.json().get("title"))

            target_keyword_hits = 0
            searchad_data = []
            result = response.json()

            for kwd in result.get("keywordList"):
                if kwd["relKeyword"] in hint_keywords_map:

                    kwd[self.MPQC] = (
                        kwd[self.MPQC] if kwd[self.MPQC] != self.LESS_THAN_10 else 9
                    )
                    kwd[self.MMQC] = (
                        kwd[self.MMQC] if kwd[self.MMQC] != self.LESS_THAN_10 else 9
                    )

                    searchad_data.append(
                        {
                            "keyword": hint_keywords_map[kwd["relKeyword"]],
                            "monthlyPcQcCnt": kwd[self.MPQC],
                            "monthlyMobileQcCnt": kwd[self.MMQC],
                        }
                    )

                    target_keyword_hits += 1
                    if target_keyword_hits == len(keywords_bundle):
                        break
            return searchad_data
        except Exception as error:
            raise Exception(error)

    def get_datalab_data(
        self, keywords_bundle: List[str], start_dt, end_dt, age_gender=False
    ) -> List[DatalabResult]:
        """
        네이버 데이터랩 API 데이터 가져오기

        API 문서: https://developers.naver.com/docs/serviceapi/datalab/search/search.md#python
        """

        def __add_results_to_datalab_data(p_headers, p_body, p_datalab_data, p_age_gender=False):
            response = requests.post(
                url=self.DATALAB_URL, headers=p_headers, data=json.dumps(p_body)
            )

            if not response.ok:
                self._datalab_key = self._get_datalab_key()

                if response.status_code == 401:
                    # ============================================================
                    # NOTE: 데이터랩 인증 실패
                    self._update_datalab_key(self._datalab_key["key_id"], "auth")
                    # ============================================================
                elif response.status_code == 429:
                    # ============================================================
                    #  NOTE: 데이터랩 호출 한도 초과시 업데이트
                    self._update_datalab_key(self._datalab_key["key_id"], "active")
                    # ============================================================

                return self.get_datalab_data(keywords_bundle, start_dt, end_dt)

            results = response.json().get("results")
            for result in results:
                keyword = result.get("title", "")
                data = result.get("data", [])
                if p_age_gender:
                    data = [d.update({"gender": gender, "age": age}) or d for d in data]

                temp_data = data.copy()
                existing_dates = {d["period"] for d in data}

                start_date = pendulum.from_format(start_dt, "YYYY-MM-DD")
                end_date = pendulum.from_format(end_dt, "YYYY-MM-DD")

                current_date = start_date
                while current_date <= end_date:
                    current_date_str = current_date.strftime("%Y-%m-%d")
                    if current_date_str not in existing_dates:
                        if p_age_gender:
                            temp_data.append({
                                "period": current_date_str, "ratio": 0.0, "gender": gender, "age": age
                            })
                        else:
                            temp_data.append({"period": current_date_str, "ratio": 0.0})
                    current_date = current_date.add(days=1)

                temp_data.sort(key=lambda x: x["period"])
                p_datalab_data.append({"keyword": keyword, "data": temp_data})

        try:
            if len(keywords_bundle) > 5:
                raise ValueError("keywords_bundle 리스트는 최대 5개까지만 허용됩니다.")

            datalab_data = []
            headers = {
                "X-Naver-Client-Id": self._datalab_key.get("client_id"),
                "X-Naver-Client-Secret": self._datalab_key.get("client_secret"),
                "Content-Type": "application/json",
            }
            body = {
                "startDate": start_dt,
                "endDate": end_dt,
                "timeUnit": "date",
                "keywordGroups": [
                    {"groupName": k, "keywords": [k]} for k in keywords_bundle
                ],
            }

            if age_gender:
                ages_category: dict = dict({str(x): [str(x)] for x in range(1, 12)})
                genders_category: dict = {"남": "m", "여": "f"}

                for age, age_value in ages_category.items():
                    for gender, gender_value in genders_category.items():
                        body.update({"gender": gender_value, "ages": age_value})
                        __add_results_to_datalab_data(headers, body, datalab_data, age_gender)
            else:
                __add_results_to_datalab_data(headers, body, datalab_data, age_gender)
            return datalab_data

        except Exception as error:
            raise Exception(error)

    def calc_search_volume(
        self,
        keywords_bundle,
        searchad_data: List[SearchadResult],
        datalab_data: List[DatalabResult],
        age_gender=False
    ):
        def __add_results_to_search_volume(p_data, p_total_ratio, p_pc_volume, p_mo_volume, p_search_volume, p_age_gender):
            total_volume = p_pc_volume + p_mo_volume
            volume_per_ratio = total_volume / p_total_ratio if p_total_ratio > 0 else 0

            for d in p_data.get("data"):
                if p_age_gender:
                    new_data = {
                        "keyword_text": kwd,
                        "keyword_date": pendulum.parse(d["period"], tz=tz),
                        "pc_volume": p_pc_volume,
                        "mo_volume": p_mo_volume,
                        "total_volume": total_volume,
                        "volume_ratio": d["ratio"],
                        "keyword_volume": int(d["ratio"] * volume_per_ratio),
                        "gender": d["gender"],
                        "age": d["age"],
                        "collected_at": pendulum.today(tz=tz),
                    }
                else:
                    new_data = {
                        "keyword_text": kwd,
                        "keyword_date": pendulum.parse(d["period"], tz=tz),
                        "pc_volume": p_pc_volume,
                        "mo_volume": p_mo_volume,
                        "total_volume": total_volume,
                        "volume_ratio": d["ratio"],
                        "keyword_volume": int(d["ratio"] * volume_per_ratio),
                        "collected_at": pendulum.today(tz=tz),
                    }
                p_search_volume.append(new_data)

        try:
            search_volume = []

            for kwd in keywords_bundle:
                tgt_searchad_datum = [d for d in searchad_data if d["keyword"] == kwd]
                tgt_datalab_datum = [d for d in datalab_data if d["keyword"] == kwd]

                if tgt_searchad_datum and tgt_datalab_datum:
                    monthly_pc_volume = tgt_searchad_datum[0].get(self.MPQC)
                    monthly_mo_volume = tgt_searchad_datum[0].get(self.MMQC)

                    if age_gender:
                        total_ratio = sum(age_gender_data["ratio"]
                                          for keyword_age_gender_data in tgt_datalab_datum
                                          for age_gender_data in keyword_age_gender_data.get("data"))
                        for datum in tgt_datalab_datum:
                            __add_results_to_search_volume(datum, total_ratio, monthly_pc_volume, monthly_mo_volume, search_volume, age_gender)
                    else:
                        total_ratio = sum([d["ratio"] for d in tgt_datalab_datum[0].get("data")])
                        __add_results_to_search_volume(
                            tgt_datalab_datum[0],
                            total_ratio,
                            monthly_pc_volume,
                            monthly_mo_volume,
                            search_volume,
                            age_gender
                        )

            return search_volume

        except Exception as error:
            logger.error(error)

    def get_search_volume(self, age_gender=False):
        try:
            results = []

            if not self._searchad_key or not self._datalab_key:
                return results

            self._set_date_range(target_date=self._date_until)

            self._set_keywords(self.get_keywords())
            self._set_keywords_bundle(keywords=self._keywords)

            for kwds_bundle in self._keywords_bundle:
                searchad_data = self.get_searchad_data(kwds_bundle)

                for date_range in self._date_range:
                    start_dt, end_dt = [d.strftime("%Y-%m-%d") for d in date_range]
                    logger.info(f"{start_dt} ~ {end_dt} {kwds_bundle}")

                    datalab_data = self.get_datalab_data(kwds_bundle, start_dt, end_dt, age_gender)
                    search_volume = self.calc_search_volume(
                        kwds_bundle,
                        searchad_data,
                        datalab_data,
                        age_gender
                    )
                    results.extend(search_volume)
            return results
        except Exception as error:
            logger.error(error)


# if __name__ == "__main__":
#     try:
#         naver_search_volume_crawler = NaverSearchVolumeCrawler()
#         search_volume = naver_search_volume_crawler.get_search_volume()
#
#         for vol in search_volume:
#             logger.debug(vol)
#     except Exception as error:
#         logger.error(error)



import time
from typing import List

import pendulum
from loguru import logger
from sqlalchemy.sql import func
from sqlalchemy import text, Column, Integer, String, Date, TIMESTAMP, UniqueConstraint
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from dmk_packages.database import database as db


class YoutubeTokenHandler:
    # ==============================================================================
    # NOTE : 한투(postgres)로 데이터베이스 설정
    def __init__(self, target, meta_table_name):
        self._db_engine = db.get_engine(target)
        self._meta_table = meta_table_name
        if not self._db_engine:
            raise ValueError("데이터베이스 엔진 설정에 실패했습니다.")

        self._create_meta_table(self._db_engine, self._meta_table)
    # ==============================================================================

    # ==============================================================================
    # NOTE : 메타테이블 생성 함수
    @staticmethod
    def _create_meta_table(engine, meta_table_name):
        try:
            db.create_to_postgres(
                engine,
                meta_table_name,
                Column("id", Integer, primary_key=True, autoincrement=True),
                Column("keyword", String),
                Column("next_page_token", String),
                Column("target_date", Date),
                Column("created_at", TIMESTAMP(timezone=False), server_default=func.timezone('KST', func.current_timestamp()), nullable=False),
                UniqueConstraint("id", name="t_metadata_youtube_pk_test")
            )
        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : json 응답에 있는 "next_page_token" 키에 해당하는 값을 metadata 테이블에 저장
    def _save_page_token(self, keyword, target_date, next_page_token):
        query = f"""
        INSERT INTO {self._meta_table} (keyword, next_page_token, target_date)
        VALUES ('{keyword}', '{next_page_token}', '{target_date}');
        """

        try:
            with self._db_engine.begin() as connection:
                connection.execute(text(query))
            logger.info(
                f"[{keyword}][{target_date}]: Metadata insertion completed."
            )
        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : 주어진 날짜, 키워드에 해당하는 next_page_token을 metadata 테이블에서 조회
    def _get_page_token(self, keyword, target_date):
        query = f"""
        SELECT next_page_token
        FROM {self._meta_table}
        WHERE keyword = '{keyword}'
            AND target_date = '{target_date}';
        """

        try:
            with self._db_engine.begin() as connection:
                result = connection.execute(text(query)).fetchone()
                return result[0] if result else False
        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : 주어진 날짜, 키워드에 해당하는 next_page_token 업데이트
    def _update_page_token(self, keyword, target_date, next_page_token):
        query = f"""
        UPDATE {self._meta_table}
        SET next_page_token = '{next_page_token}'
        WHERE keyword = '{keyword}'
            AND target_date = '{target_date}'
        """

        try:
            with self._db_engine.begin() as connection:
                connection.execute(text(query))
            logger.info(f"[{keyword}][{target_date}]: Metadata update completed.")
        except Exception as err:
            logger.error(err)
    # ==============================================================================


class YoutubeKeyHandler:
    # ==============================================================================
    # NOTE : 마대리(postgres)로 데이터베이스 설정
    def __init__(self, target):
        self._maderi_engine = db.get_engine(target)
        self._api_key_table = 't_auth_yt_api_key_v2'
        if not self._maderi_engine:
            raise ValueError("데이터베이스 엔진 설정에 실패했습니다.")

        self.__reset_youtube_apikey()
        self._youtube_api_key = self._get_youtube_apikey()
    # ==============================================================================

    # ==============================================================================
    # NOTE : 마대리에 저장되어 있는 유튜브 API 키 리셋
    def __reset_youtube_apikey(self):
        query = f"""
        UPDATE {self._api_key_table}
        SET is_active = TRUE,
            is_valid = TRUE,
            failed_active_date = NULL,
            failed_valid_date = NULL
--         WHERE NOT is_active AND failed_active_date < CURRENT_DATE
        """
        try:
            with self._maderi_engine.begin() as connection:
                connection.execute(text(query))
            logger.info("API 키 리셋 완료")
        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : 사용 가능한 유튜브 API 키 랜덤으로 하나 가져오기
    def _get_youtube_apikey(self):
        query = f"""
        SELECT id, api_key
        FROM {self._api_key_table}
        WHERE is_valid AND is_active 
        ORDER BY RANDOM()
        LIMIT 1
        """

        try:
            with self._maderi_engine.begin() as connection:
                result = connection.execute(text(query))
            youtube_key = result.fetchone()
            if not youtube_key:
                raise ValueError("사용 가능한 유튜브 API키가 없습니다")
            return {
                "key_id": youtube_key[0],
                "api_key": youtube_key[1]
            }
        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : API 키가 할당량을 채웠을 경우 사용 불가하다 표시
    def _update_youtube_apikey_state(self, pk_id, feather):
        query = f"""
        UPDATE {self._api_key_table}
        SET is_{feather} = FALSE, 
            failed_{feather}_date = NOW()::timestamp
        WHERE id = {pk_id}
        """
        try:
            if self._maderi_engine is None:
                raise Exception("인증 정보 DB에 접근할 수 없습니다.")

            if feather not in ["valid", "active"]:
                raise Exception("사용할 수 없는 feather입니다.")

            with self._maderi_engine.begin() as connection:
                connection.execute(text(query))
        except Exception as error:
            raise Exception(error)
    # ==============================================================================

    # ==============================================================================
    # NOTE : API 키 사용 불가할 경우, 해당 키의 상태 업데이트 & 사용 가능한 키로 변경
    def _change_youtube_apikey(self, err):
        self._api_key = self._get_youtube_apikey()

        if err.resp.status == 400:
            self._update_youtube_apikey_state(self._api_key["key_id"], "valid")
        elif err.resp.status == 403:
            self._update_youtube_apikey_state(self._api_key["key_id"], "active")
    # ==============================================================================


class YoutubeCrawler(YoutubeKeyHandler, YoutubeTokenHandler):
    tz = pendulum.timezone("Asia/Seoul")
    DEFAULT_DT = pendulum.yesterday(tz=tz)

    def __init__(self, key_auth, token_auth, meta_name):
        YoutubeKeyHandler.__init__(self, target=key_auth)
        YoutubeTokenHandler.__init__(self, target=token_auth, meta_table_name=meta_name)

        self._keywords: List[str] | None = None
        self._api_key = self._get_youtube_apikey()
        self._results = []

    @staticmethod
    def _kst_to_utc(kst_string):
        kst_datetime = pendulum.parse(kst_string, tz="Asia/Seoul")
        utc_datetime = kst_datetime.in_timezone("UTC")
        return utc_datetime.isoformat()

    @staticmethod
    def _utc_to_kst(utc_string):
        utc_datetime = pendulum.parse(utc_string, tz="UTC")
        kst_datetime = utc_datetime.in_timezone("Asia/Seoul")
        return kst_datetime.isoformat()

    def get_keywords(self) -> List[str]:
        """해당 메서드를 오버라이딩하여 키워드를 세팅합니다."""

        # NOTE: 예시를 위한 키워드입니다.
        return ["Python", "Rust", "JavaScript", "Java", "Flutter"]

    # ==============================================================================
    # NOTE : 하나의 영상에 대한 정보 반환
    def _get_video_info(self, keyword, video_info):
        try:
            youtube = build(
                serviceName="youtube",
                version="v3",
                developerKey=self._api_key["api_key"]
            )
            response = youtube.videos().list(
                id=video_info["id"]["videoId"],
                part="snippet,statistics"
            ).execute()
            time.sleep(0.1)

            video = response.get("items")[0] if len(response.get("items")) > 0 else {}
            video_id = video_info.get("id", {}).get("videoId")
            video_snippet = video.get("snippet", {})
            video_statistics = video.get("statistics", {})
            result = {
                "title": video_snippet.get("title", ""),
                "creator": video_snippet.get("channelTitle", ""),
                "contents": video_snippet.get("description", "").replace("\n\n", "").replace("\n", " "),
                "regist_date": self._utc_to_kst(video_snippet.get("publishedAt")),
                "view": int(video_statistics.get("viewCount", 0)),
                "recommend": int(video_statistics.get("likeCount", 0)),
                "comment": int(video_statistics.get("commentCount", 0)),
                "url": f"https://www.youtube.com/watch?v={video_id}",
                "keyword": keyword
            }
            return result

        except HttpError as err:
            key_id = self._api_key['key_id']
            self._change_youtube_apikey(err)

            logger.info(f"{err.resp.status} {err.resp.reason} 에러 발생 > {key_id}번 API 키로 변경")
            return self._get_video_info(keyword, video_info)

        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : 주어진 날짜, 키워드에 대한 전체 결과를 _results 리스트에 추가 후 반환
    def _stack_videos_info(self, keyword, target_date, page_token=None):
        try:
            next_page_token = page_token
            published_after = self._kst_to_utc(f"{target_date} 00:00:00").replace("+00:00", "Z")
            published_before = self._kst_to_utc(f"{target_date} 23:59:59").replace("+00:00", "Z")
            youtube = build(
                serviceName="youtube",
                version="v3",
                developerKey=self._api_key["api_key"]
            )
            response = youtube.search().list(
                q=keyword,
                type="video",
                part="id,snippet",
                maxResults=50,  # NOTE: 최대 50개
                pageToken=next_page_token,
                publishedAfter=published_after,
                publishedBefore=published_before,
                order="date",
                regionCode="KR",
            ).execute()
            time.sleep(0.1)

            videos = response.get("items", [])
            next_page_token = response.get("nextPageToken")

            # 일부 지정한 시간대 범위에서 벗어난 업로드 시간의 영상이 있음 -> 해당 영상은 걸러내기
            results = [self._get_video_info(keyword, video) for video in videos
                       if published_after <= video["snippet"]["publishedAt"] <= published_before]
            self._results.extend(results)
            logger.info(f"[{keyword}][{target_date}]: {len(results)} 개 추가 수집")

            if not page_token:
                self._save_page_token(keyword, target_date, next_page_token)
            else:
                self._update_page_token(keyword, target_date, next_page_token)

            if next_page_token:
                self._stack_videos_info(keyword, target_date, next_page_token)
            else:
                logger.info(f"[{keyword}][{target_date}]: {len(self._results)} 개의 영상 정보 수집 완료")

        except HttpError as err:
            key_id = self._api_key['key_id']
            self._change_youtube_apikey(err)

            logger.info(f"{err.resp.status} {err.resp.reason} 에러 발생 > {key_id}번 API 키로 변경")
            return self._stack_videos_info(keyword, target_date, next_page_token)

        except Exception as err:
            logger.error(err)
    # ==============================================================================

    # ==============================================================================
    # NOTE : _results 값 반환
    def get_videos_info(self, keyword, target_date):
        next_page_token = self._get_page_token(keyword, target_date)

        if next_page_token == 'None':
            logger.info(f"[{keyword}][{target_date}]: 이미 수집 완료")
        elif not next_page_token:
            # logger.info(f"[{keyword}][{target_date}]: 수집 필요")
            self._stack_videos_info(keyword, target_date)
        else:
            logger.info(f"[{keyword}][{target_date}]: next_page_token 업데이트 필요")
            self._update_page_token(keyword, target_date, next_page_token)

        results = self._results.copy()
        self._results = []
        return results
    # ==============================================================================



import json
from typing import TypedDict
from urllib.parse import urlencode

import requests
from fake_useragent import UserAgent
from loguru import logger

BlogItem = TypedDict(
    "BlogItem",
    {
        "domainIdOrBlogId": str,
        "logNo": int,
        "gdid": str,
        "postUrl": str,
        "title": str,
        "noTagTitle": str,
        "contents": str,
        "nickName": str,
        "blogName": str,
        "profileImgUrl": str,
        "addDate": int,
        "thumbnails": list[str],
        "product": None,  # FIXME: 다른 데이터는 어떤 것인지 파악되지 않음
        "marketPost": bool,
        "hasThumbnail": bool,
    },
)


class NaverBlogCrawler:
    BASE_URL = "https://section.blog.naver.com/ajax/SearchList.naver"
    REFERER_URL = "https://section.blog.naver.com/Search/Post.naver"

    def __init__(self):
        self._user_agent = UserAgent(min_percentage=1.1, os=["windows", "macos"])

    def get_blog_items(self, keyword: str, page: int, start_dt: str, end_dt: str):
        base_url_params = {
            "keyword": keyword,
            "currentPage": page,
            "startDate": start_dt,
            "endDate": end_dt,
            "orderBy": "recentdate",
            "countPerPage": "15",
            "type": "post",
        }

        referer_url_params = {
            "keyword": keyword,
            "startDate": start_dt,
            "endDate": end_dt,
            "orderBy": "recentdate",
            "rangeType": "PERIOD",
            "pageNo": "1",
        }

        headers = {
            "User-agent": self._user_agent.random,
            "Referer": f"{self.REFERER_URL}?{urlencode(referer_url_params)}",
        }

        response = requests.get(
            f"{self.BASE_URL}?{urlencode(base_url_params)}", headers=headers
        )

        data = json.loads(response.text.split(",", 1)[-1].strip())

        total_count = data.get("result", {}).get("totalCount", [])
        search_list = data.get("result", {}).get("searchList", [])

        logger.debug(search_list[0])

        return {"total_count": total_count, "search_list": search_list}

    def parse_blog_item(self): ...


if __name__ == "__main__":
    naver_blog_crawler = NaverBlogCrawler()

    naver_blog_crawler.get_blog_items("파이썬", 1, "2024-03-19", "2024-03-19")



import os
import requests
import xml.etree.ElementTree as ET
from datetime import datetime
from dotenv import load_dotenv, find_dotenv

load_dotenv(dotenv_path=find_dotenv())


def check_day_off():
    today = datetime.now()
    year = datetime.now().strftime("%Y")
    month = datetime.now().strftime("%m")

    url = (
        "http://apis.data.go.kr/B090041/openapi/service/SpcdeInfoService/getRestDeInfo"
    )
    serviceKey = os.environ.get("DAY_OFF_API_KEY")
    params = {"solYear": f"{year}", "solMonth": f"{month}", "serviceKey": serviceKey}

    response = requests.get(url, params=params)

    root = ET.fromstring(response.text)
    dates = [item.find("locdate").text for item in root.findall(".//item")]

    holiday = []
    for date in dates:
        holiday.append(date)

    if today.strftime("%Y%m%d") in holiday:
        return True

    elif str(today.weekday()) == "5" or str(today.weekday()) == "6":
        return True

    else:
        return False



import re
import time
import pendulum
import requests
from bs4 import BeautifulSoup
from fake_useragent import UserAgent
from loguru import logger
import random
from sqlalchemy import MetaData, Table, select, func

"""
## Clien 크롤러 ##
- 사용 유의사항
카테고리와 검색할 키워드를 묶어서 매개변수로 전달해줘야함
ex. targets = [
                (keyword, category)
                for category in self._cat_list
                for keyword in keywords
            ]
            for target in targets:
                keyword, category = target

- 사용 순서
check_borad_sn > start_beautifulsoup > crawling_posts > check_next_page > crawling_post_detail_list
> 디비 적재 > 페이지가 더 있다면 처음부터 다시 시작(페이지 추가)

"""


class ClienCrawler:
    DEFAULT_DT = pendulum.yesterday(tz=pendulum.timezone("Asia/Seoul"))

    def __init__(self, date_from=None, date_until=None):

        self._user_agent = UserAgent(min_percentage=1.1, os=["windows", "macos"])
        self._session = requests.session()
        self._max_board_sn = 0
        self._min_board_sn = 0
        self._base_url = "https://www.clien.net"

        self._date_from = pendulum.parse(date_from) if date_from else self.DEFAULT_DT
        _date_until = pendulum.parse(date_until) if date_until else self.DEFAULT_DT
        _date_until = _date_until.strftime("%Y-%m-%d 23:59:59")
        self._date_until = pendulum.parse(_date_until)

    def check_board_sn(self, keyword, category, table, engine):
        """
        keyword 및 category에 해당하는 board_sn컬럼의 max, min 값을 가져오는 컬럼
        """

        metadata = MetaData()
        metadata.bind = engine
        table = Table(table, metadata, autoload_with=engine)

        stmt = select(func.max(table.c.board_sn), func.min(table.c.board_sn)).where(
            (table.c.category_name == category) & (table.c.keyword == keyword)
        )

        with engine.begin() as connection:
            result = connection.execute(stmt)
            a = result.fetchall()
            max_board_sn = a[0][0] or 0
            min_board_sn = a[0][1] or 0

        return max_board_sn, min_board_sn

    def start_beautifulsoup(self, keyword: str, category: str, page: int):
        headers = {"User-agent": self._user_agent.random}
        target_url = self._base_url + "/service/search"
        keyword_r = keyword.replace("&", "%26").replace("(", "").replace(")", "")
        params = {
            "q": keyword_r,
            "p": page,
            "sort": "recency",
            "boardCd": category,
            "isBoard": True,
        }

        response = self._session.get(
            target_url, headers=headers, params=params, timeout=3
        )
        time.sleep(random.uniform(1, 10))

        soup = BeautifulSoup(response.text, "html.parser")
        return soup

    def crawling_posts(self, soup: BeautifulSoup, keyword: str, category: str):
        """
        게시물들의 기본적인 내용 크롤링
        """
        items = soup.select(".list_item.symph_row.jirum")
        results = []

        for item in items:
            board_sn = int(item["data-board-sn"])
            regist_date = pendulum.parse(item.select_one(".timestamp").text)

            title = item.select_one(".subject_fixed").text
            if "헤드라인 모음" in title:
                continue

            if self._min_board_sn <= board_sn <= self._max_board_sn:
                continue

            if self._date_from < regist_date < self._date_until:
                comment = int(item["data-comment-count"])
                url = self._base_url + item.select_one(".subject_fixed")["href"]

                data = {
                    "category_name": category,
                    "regist_date": regist_date.to_iso8601_string(),
                    "comment": comment,
                    "url": url,
                    "keyword": keyword,
                    "board_sn": board_sn,
                }
                results.append(data)
        return results

    def check_next_page(self, soup: BeautifulSoup):
        pages = soup.select(".board-nav-page")
        if len(pages) <= 1:
            return None
        next_btn_exists = bool(soup.select_one(".board-nav-next"))

        last_item = soup.select(".list_item.symph_row.jirum")[-1]
        last_item_timestamp = pendulum.parse(last_item.select_one(".timestamp").text)

        if self._date_from > last_item_timestamp:
            return None

        for idx, page in enumerate(pages):
            if "active" in page.get("class", []):
                if next_btn_exists or idx + 1 < len(pages):
                    return idx + 1
                break
        return None

    def crawling_post_detail_list(self, posts):
        """
        해당 게시물의 디테일한 내용 가져와서 리스트에 저장
        """
        results = []
        for post in posts:
            logger.info(f"[적재하는 url: {post['url']}")

            details = self._crawling_post_detail(post.get("url"))
            results.append({**post, **details})
        return results

    def _crawling_post_detail(self, post_url):
        """
        해당 게시물의 디테일한 내용 정제하기
        """
        headers = {"User-agent": self._user_agent.random}
        response = self._session.get(post_url, headers=headers)
        time.sleep(random.uniform(1, 10))

        soup = BeautifulSoup(response.text, "html.parser")

        title_elems = soup.select(".post_subject span")
        title = next(
            (
                elem.text.strip()
                for elem in title_elems
                if "post_category" not in elem.get("class", [])
            ),
            None,
        )

        view = soup.select_one(".view_count strong").text.strip().replace(",", "")

        contents = soup.select_one(".post_article").text.strip()
        # 연속된 두 개 이상의 개행을 한 개의 개행으로 줄임
        contents = re.sub(r"\n\n", "\n", contents)
        # 개행 문자를 공백으로 치환
        contents = re.sub(r"\n", " ", contents)
        # \xa0, \xad, \ufeff 문자 제거
        contents = re.sub(r"[\xa0\xad\ufeff]+", "", contents)
        # 두 개 이상의 공백을 단일 공백으로 치환
        contents = re.sub(r" {2,}", " ", contents)

        return {"view": int(view), "title": title, "contents": contents}



import os
import asyncio
import pendulum
from dotenv import load_dotenv, find_dotenv
from playwright.async_api import async_playwright

"""
## bigkinds 크롤러 ##
- 사용 유의사항
해당 크롤러는 bigkinds의 분석결과까지 도달함. 이후에 분석결과의 5가지 종류중 필요한 데이터를 찾아서 다운로드

- 사용 순서
start_playwright > crawling_bigkinds > 원하는 데이터 다운로드 및 디비 적재 > stop_playwright
"""
load_dotenv(dotenv_path=find_dotenv())


class BigkindsCrawler:
    DEFAULT_DT = pendulum.yesterday(tz=pendulum.timezone("Asia/Seoul"))

    def __init__(self, date_from=None, date_until=None, login_env=None):

        _date_from = pendulum.parse(date_from) if date_from else self.DEFAULT_DT
        _date_until = pendulum.parse(date_until) if date_until else self.DEFAULT_DT
        self._date_from = _date_from.to_date_string()
        self._date_until = _date_until.to_date_string()

        self.login_id = os.environ.get(f"{login_env}_LOGIN_ID")
        self.login_pw = os.environ.get(f"{login_env}_LOGIN_PW")
        if None in (self.login_id, self.login_pw):
            raise Exception("check your login in .env")

    async def _start_playwright(self, view=True):
        self.playwright = await async_playwright().start()
        self.browser = await self.playwright.chromium.launch(headless=view)
        self.context = await self.browser.new_context(accept_downloads=True)
        self.page = await self.context.new_page()
        return self

    async def stop_playwright_async(self):
        await self.context.close()
        await self.browser.close()
        await self.playwright.stop()

    async def crawling_bigkinds_async(self, cat, view=True):
        """
        사이트 접속후 분석결과 페이지까지 도달
        """
        # playwright 실행
        await self._start_playwright(view)

        # 사이트 접속
        url = "https://www.bigkinds.or.kr/v2/news/index.do"
        await self.page.goto(url, wait_until="load")
        await asyncio.sleep(2)

        # 로그인
        await self.page.get_by_role("button", name="로그인").click()
        await asyncio.sleep(0.5)

        await self.page.get_by_placeholder("이메일(E-mail) 주소").fill(self.login_id)
        await self.page.get_by_placeholder("비밀번호를 입력하세요.").fill(self.login_pw)
        await self.page.click("#login-btn")
        await asyncio.sleep(2)

        await self.page.wait_for_selector(
            ".ft-map > div > div > ul > li > a", state="visible"
        )
        await self.page.click(".ft-map > div > div > ul > li > a")
        await asyncio.sleep(2)

        # 기간설정하기
        await self.page.click(".tab-btn")
        date_s = await self.page.query_selector("#search-begin-date")
        date_e = await self.page.query_selector("#search-end-date")
        await asyncio.sleep(0.3)

        await date_s.fill(self._date_from)
        await date_e.fill(self._date_until)

        # 해당되는 카테고리 설정
        await self.page.click(".tab-btn.tab3")
        set_cat = await self.page.query_selector(
            f"xpath=.//span[@data-role='display' and text()='{cat}']"
        )
        await set_cat.click()
        await self.page.get_by_role("button", name="적용하기").click()
        await asyncio.sleep(2)

        # STEP 03 분석결과 버튼 누르기
        await self.page.get_by_role("button", name="분석 결과 및 시각화").click()
        await asyncio.sleep(1)



import os
import random
import asyncio
import pendulum
from loguru import logger
from playwright.async_api import (
    async_playwright,
    TimeoutError as PlaywrightTimeoutError,
)
from dotenv import load_dotenv, find_dotenv

"""
## fnguide 크롤러 ##
- 사용 유의사항
해당 크롤러는 가져올 데이터는 일정하지만 카테고리가 상당히 많기때문에(서브카테고리 포함) 필요한 카테고리 elements를 매개변수로 지정
ex. click_cat = (f'//*[@id="resultDivTabs"]/ul/li/button[contains(text(), "{cat}")]')

download경로를 None으로 설정할시 pdf파일은 다운로드 하지 않음

- 사용 : run_fnguide_crawler(카테고리, clickelements, download_path, view)
"""
load_dotenv(dotenv_path=find_dotenv())


class FnguideCrawler:
    DEFAULT_DT = pendulum.yesterday(tz=pendulum.timezone("Asia/Seoul"))

    def __init__(self, date_from=None, date_until=None, login=None):

        _date_from = pendulum.parse(date_from) if date_from else self.DEFAULT_DT
        _date_until = pendulum.parse(date_until) if date_until else self.DEFAULT_DT
        self._date_from = _date_from.to_date_string()
        self._date_until = _date_until.to_date_string()

        self.login_id = os.environ.get(f"{login}_LOGIN_ID")
        self.login_pw = os.environ.get(f"{login}_LOGIN_PW")
        if None in (self.login_id, self.login_pw):
            raise Exception("check your login in .env")

        # 다운로드 에러가 가는 url 리스트
        self.url_e = []

    async def _start_playwright(self, view=True):
        self.playwright = await async_playwright().start()
        self.browser = await self.playwright.chromium.launch(
            headless=view  # 창띄울지 말지 결정
        )
        self.context = await self.browser.new_context(accept_downloads=True)
        self.page = await self.context.new_page()
        return self

    async def _stop_playwright(self):
        await self.context.close()
        await self.browser.close()
        await self.playwright.stop()

    async def _set_login_date(self):
        """
        사이트 접속후 로그인 및 기간설정
        """

        # 로그인
        url_fnguide = "https://www.fnguide.com/"
        await self.page.goto(url_fnguide, wait_until="load")
        await asyncio.sleep(2)

        await self.page.fill('input[name="MemberID"]', self.login_id)
        await self.page.fill('input[name="PassWord"]', self.login_pw)
        await self.page.click(".btn--login")
        await asyncio.sleep(2)

        # fnresearch 들어가서 기간설정 후 검색
        await self.page.click(".gnb--1dep > .p1")
        await asyncio.sleep(2)

        await self.page.fill("#frDate", self._date_from)
        await self.page.fill("#toDate", self._date_until)
        await self.page.locator("#srchBtn").click()
        await asyncio.sleep(0.5)

    async def _crawling_fnguide(self, cat, click_cat, download_path):
        """
        해당 카테고리 클릭후 크롤링 및 PDF파일 다운로드
        """
        # 카테고리 클릭
        await self.page.click(click_cat)
        await asyncio.sleep(0.5)

        data_list = []
        while True:
            # 검색결과가 없는 경우
            table = await self.page.query_selector("#resultDivGrid")
            if await table.query_selector(".nodata"):
                break

            rows = await table.query_selector_all("tbody > tr")
            for row in rows:
                tds = await row.query_selector_all("td")
                regist_date = await tds[0].inner_text()
                category_name = cat
                title = await tds[1].inner_text()
                provider = await tds[4].inner_text()
                writer = await tds[3].inner_text()
                url_tag = await tds[1].eval_on_selector(
                    "a", 'element => element.getAttribute("href")'
                )
                url = f"https://www.fnguide.com{url_tag}"

                # 찾은 데이터 list에 append
                data = {
                    "channel_name": "Fnguide",
                    "category_name": category_name,
                    "regist_date": regist_date,
                    "title": title,
                    "writer": writer,
                    "provider": provider,
                    "url": url,
                }
                data_list.append(data)

                # pdf 다운로드
                try:
                    if download_path is None:
                        continue

                    # 이미 다운로드 되어있는 파일 목록
                    url_idx = url.split("bulletkind=")[-1]
                    filename = f"{url_idx}_{regist_date}.pdf"

                    os.chdir(download_path)
                    file_list = os.listdir()

                    # 오류나는 파일과 이미 다운로드가 되어있다면 패스 아니라면 다운로드 진행
                    if (url_tag not in self.url_e) and (filename not in file_list):
                        async with self.page.expect_download(
                            timeout=15000
                        ) as download_info:
                            pdf = await row.query_selector(".btn--get")
                            await pdf.click()

                        download = await download_info.value
                        await download.save_as(download_path + filename)
                        await asyncio.sleep(random.uniform(45,65))

                except PlaywrightTimeoutError:
                    # 오류난 파일의 url_tag 추가
                    logger.error(f"[{cat}][{regist_date}][{title}] download error")
                    self.url_e.append(url_tag)

                    # 다시 실행
                    logger.info(f"[{cat}] rerun")
                    await self._stop_playwright
                    await self.run_fnguide_crawler_async(cat, click_cat, download_path)

            # 반목문 중단 or 다음 페이지
            try:
                await self.page.click(".paging > .btn--next", timeout=1000)
                await asyncio.sleep(0.5)

            except PlaywrightTimeoutError:
                break

        return data_list

    async def run_fnguide_crawler_async(self, cat, click_cat, download_path, view=True):
        """
        fnguide 크롤러 전체적인 실행후 데이터 반환
        """
        await self._start_playwright(view)
        await self._set_login_date()
        data = await self._crawling_fnguide(cat, click_cat, download_path)
        await asyncio.sleep(random.uniform(300,600))
        await self._stop_playwright()
        return data
